# frozen_string_literal: true

module FlakeySpecCatcher
  # CapsuleManager class
  #
  # Contains a file by file summary of all git changes.
  #
  # A CapsuleManager object contains all ChangeCapsule objects. It delivers
  # summaries of all changes and runs methods on the contained ChangeCapsule
  # objects.
  class CapsuleManager
    attr_reader :change_capsules

    def initialize
      @change_capsules = []
    end

    def add_capsule(capsule)
      @change_capsules.push(capsule)
    end

    def changed_examples
      @change_capsules.map(&:changed_examples).flatten.uniq
    end

    def changed_files
      @change_capsules.map(&:file_name).uniq
    end

    def sorted_change_contexts
      @change_capsules.map(&:change_contexts).flatten.sort_by { |c| c.line_number.to_i }
    end

    def condense_reruns
      # Don't re-run if the parent context of a change will be run
      reruns = []
      all_contexts = sorted_change_contexts
      all_contexts.each do |context|
        next if reruns.include?(context.file_name) || context.ancestor_present_in_reruns(reruns)

        reruns.push(context.rerun_info) unless reruns.include?(context.rerun_info)
      end
      reruns
    end

    def tag_values_equal?(user_excluded_tag_value, rspec_tag_value)
      return true if user_excluded_tag_value.nil? && rspec_tag_value.nil?

      return false if user_excluded_tag_value.nil? || rspec_tag_value.nil?

      (user_excluded_tag_value.tr('\'\"', '') == rspec_tag_value.tr('\'\" ', ''))
    end

    def find_reruns_by_tags(user_excluded_tags)
      dropped_tests = []
      sorted_change_contexts.each do |context|
        context.tags.keys.each do |tag|
          next unless user_excluded_tags.key?(tag) &&
                      tag_values_equal?(user_excluded_tags[tag], context.tags[tag])

          dropped_tests << context.rerun_info
        end
      end
      dropped_tests
    end
  end
end
