# frozen_string_literal: true

require_relative '../helpers/colorize'
require_relative '../helpers/indent_string'

module FlakeySpecCatcher
  # RspecResult class
  #
  # Organizes results for a changed test (spec).
  #
  # Each changed test will have one RspecResult created for it and for each
  # re-run of that spec, the results will be pushed to the RspecResult object.
  # This class will then organize and output the results accordingly.
  class RspecResult
    attr_accessor :description, :location, :total_times_run, :total_failures

    def initialize(description, location, exception_message = nil)
      @description = description
      @location = location
      @total_times_run = 1
      @total_failures = exception_message ? 1 : 0
      @failures = []
      add_failure(exception_message) if exception_message
    end

    def add_run(exception_message = nil)
      @total_times_run += 1

      return unless exception_message

      @total_failures += 1
      add_failure(exception_message)
    end

    def add_failure(exception_message)
      failure = @failures.find { |f| f.exception_message == exception_message }
      failure ? failure.add_failure : @failures.push(RSpecFailure.new(exception_message))
    end

    def print_results
      puts "\n#{@description.yellow} (#{location})
            \nFAILED #{total_failures} / #{total_times_run} times"

      @failures.each do |f|
        puts "#{f.count.to_s.indent(2)} times with exception message:"
        puts f.exception_message.indent(4).red.to_s
      end
    end

    # Simple class to contain failed example data
    class RSpecFailure
      attr_reader :exception_message, :count

      def initialize(exception_message)
        @exception_message = exception_message
        @count = 1
      end

      def add_failure
        @count += 1
      end
    end
  end
end
