# frozen_string_literal: true

require_relative '../helpers/colorize'
require_relative './rspec_result'

module FlakeySpecCatcher
  # RspecResultManager class
  #
  # Holds a collection of RSpec results and provides helper methods
  #
  # An RspecResultManager will hold a collection of results, one for each
  # distinct example. It also provides helpers for adding new results,
  # displaying aggregate results, and checking the state of the collection.
  class RspecResultManager
    def initialize(rspec_result_class)
      @result_class = rspec_result_class
      @results = []
    end

    def add_result(desc, location, message = nil)
      result = @results.find { |r| r.location == location }
      result ? result.add_run(message) : @results.push(@result_class.new(desc, location, message))
    end

    def print_results
      puts "\n********** SUMMARY **********"
      print_successes_count(successes)
      failures.each(&:print_results) if failures.any?
    end

    def failures
      @results.select { |r| r.total_failures.positive? }
    end

    def successes
      @results.select { |r| r.total_failures.zero? }
    end

    def empty?
      @results.empty?
    end

    private

    def print_successes_count(successes)
      return unless successes.any?

      rerun_count = successes.first.total_times_run
      msg = "#{successes.count} example(s) ran #{rerun_count} times without any failures"
      puts msg.green
      successes.each { |s| puts " - #{s.location}".green }
    end
  end
end
