# frozen_string_literal: true

require_relative './cli_override'
module FlakeySpecCatcher
  # UserConfig class
  #
  # Captures user-defined settings to configure RSpec re-run settings.

  class UserConfig
    attr_reader :repeat_factor, :ignore_files, :ignore_branches, :silent_mode
    attr_reader :rerun_file_only, :rspec_usage_patterns, :excluded_tags
    attr_reader :manual_rerun_patterns, :manual_rerun_usage
    attr_reader :enable_runs, :output_file, :use_parent, :dry_run
    attr_reader :split_nodes, :split_index, :verbose, :test_options
    attr_reader :break_on_first_failure

    USER_CONFIG_ENV_VARS = %w[FSC_REPEAT_FACTOR FSC_IGNORE_FILES FSC_IGNORE_BRANCHES
                              FSC_SILENT_MODE FSC_RERUN_FILE_ONLY FSC_USAGE_PATTERNS
                              FSC_EXCLUDED_TAGS FSC_OUTPUT_FILE].freeze

    def initialize(cli_override: CliOverride.new)
      apply_env_var_settings
      @cli_override = cli_override
      override_settings
    end

    private

    # rubocop:disable Metrics/AbcSize
    def apply_env_var_settings
      @repeat_factor = initialize_repeat_factor(ENV['FSC_REPEAT_FACTOR'])
      @ignore_files = env_var_string_to_array(ENV['FSC_IGNORE_FILES'])
      @ignore_branches = env_var_string_to_array(ENV['FSC_IGNORE_BRANCHES'])
      @silent_mode = env_var_string_to_bool(ENV['FSC_SILENT_MODE'])
      @rerun_file_only = env_var_string_to_bool(ENV['FSC_RERUN_FILE_ONLY'])
      @rspec_usage_patterns = env_var_string_to_pairs(ENV['FSC_USAGE_PATTERNS'])
      @excluded_tags = env_var_string_to_tags(ENV['FSC_EXCLUDED_TAGS'])
      @output_file = ENV['FSC_OUTPUT_FILE']
      @split_nodes = env_var_string_to_int_or_nil('FSC_NODE_TOTAL')
      @split_index = env_var_string_to_int_or_nil('FSC_NODE_INDEX')
    end

    def apply_cli_override
      @manual_rerun_patterns = @cli_override.rerun_patterns
      @manual_rerun_usage = @cli_override.rerun_usage
      @use_parent = @cli_override.use_parent
      @repeat_factor = @cli_override.repeat_factor if @cli_override.repeat_factor.to_i.positive?
      @break_on_first_failure = @cli_override.break_on_first_failure
      @enable_runs = @cli_override.enable_runs
      @dry_run = @cli_override.dry_run
      @split_nodes = @cli_override.split_nodes unless @cli_override.split_nodes.nil?
      @split_index = @cli_override.split_index unless @cli_override.split_index.nil?
      @excluded_tags = if @cli_override.excluded_tags.empty?
        @excluded_tags
      else
        @cli_override.excluded_tags
      end
      @output_file = set_output_file
      @verbose = @cli_override.verbose
      @test_options = @cli_override.test_options
    end
    # rubocop:enable Metrics/AbcSize

    def override_settings
      apply_cli_override
      return unless @manual_rerun_patterns.nil?

      parse_commit_message
    end

    def set_output_file
      if !@cli_override.output_file.nil?
        @cli_override.output_file
      elsif @output_file.nil? || @output_file.strip.empty?
        File::NULL
      else
        @output_file
      end
    end

    def remove_formatter_quotes(env_var)
      env_var.tr('\'\"', '')
    end

    def initialize_repeat_factor(env_var)
      env_var.to_i.positive? ? env_var.to_i : 20
    end

    def env_var_string_to_int_or_nil(env_var_name)
      env_var = ENV[env_var_name]
      if env_var.nil? || env_var.empty?
        nil
      else
        check = env_var.to_i
        raise ArgumentError, "#{env_var_name} must be positive" if check.negative?

        check
      end
    end

    def env_var_string_to_array(env_var)
      if env_var.nil? || env_var.empty?
        []
      else
        env_var.split(',').map(&:strip)
      end
    end

    def env_var_string_to_bool(env_var)
      if env_var.to_s.casecmp('true').zero?
        true
      else
        false
      end
    end

    def env_var_string_to_pairs(env_var)
      if env_var.nil? || env_var.empty?
        []
      else
        env_var.scan(/{.*?}/).map do |pair|
          pair.tr('{}', '').split('=>').map(&:strip)
        end
      end
    end

    def env_var_string_to_tags(tag_string)
      if tag_string.nil? || tag_string.empty?
        {}
      else
        tag_strings = tag_string.tr('\'\" ', '').split(',').uniq
        tags_with_values = {}
        return if tag_strings.nil? || tag_strings.empty?

        # Since tags can have a value represented as strings or symbols
        #  we'll only remove the hash rockets and not colons
        #  Example: ":tag => 'special'" => tags_with_values[:tag] = 'special'
        #  Example: ":tag => :special" => tags_with_values[:tag] = :special

        tag_strings.each do |str|
          tag_and_value = str.sub(/=>/, ' ').split(' ')
          tags_with_values[tag_and_value[0]] = tag_and_value[1]
        end

        tags_with_values
      end
    end

    def format_regex_scan_results(matches)
      # Scanning the commit message will result in an array of matches
      #  based on the specified regex. If the pattern uses groups like ours does
      #  then an array of arrays will be returned.
      # See: https://apidock.com/ruby/String/scan

      # For each array of matches, flatten the individual match group sub-array
      #  then remove the formatter quotes we require the values to be wrapped in
      matches.map { |m| remove_formatter_quotes(m.join('')) }
    end

    def parse_commit_message
      commit_message = `git show -q`
      USER_CONFIG_ENV_VARS.each do |env_var|
        matches = commit_message.scan(/#{env_var}\s*=\s*('.*?')/)
        next unless matches.count.positive?

        # In the case of multiple overrides being specified for the same config
        #  variable, we'll only use the first
        env_value = format_regex_scan_results(matches).first
        override_user_config(env_var, env_value)
      end
    end

    # rubocop:disable Metrics/CyclomaticComplexity
    def override_user_config(env_var, env_value)
      case env_var
      when 'FSC_REPEAT_FACTOR'
        @repeat_factor = initialize_repeat_factor(env_value)
      when 'FSC_IGNORE_FILES'
        @ignore_files = env_var_string_to_array(env_value)
      when 'FSC_IGNORE_BRANCHES'
        @ignore_branches = env_var_string_to_array(env_value)
      when 'FSC_SILENT_MODE'
        @silent_mode = env_var_string_to_bool(env_value)
      when 'FSC_RERUN_FILE_ONLY'
        @rerun_file_only = env_var_string_to_bool(env_value)
      when 'FSC_USAGE_PATTERNS'
        @rspec_usage_patterns = env_var_string_to_pairs(env_value)
      when 'FSC_EXCLUDED_TAGS'
        @excluded_tags = env_var_string_to_tags(env_value)
      when 'FSC_OUTPUT_FILE'
        @output_file = env_value
      end
    end
    # rubocop:enable Metrics/CyclomaticComplexity
  end
end
